package be.jeffcheasey88.peeratcode.parser.java;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.lang.reflect.Modifier;
import java.util.function.BiConsumer;
import java.util.function.BiFunction;
import java.util.function.Function;

import be.jeffcheasey88.peeratcode.parser.Bag;
import be.jeffcheasey88.peeratcode.parser.Parser;
import be.jeffcheasey88.peeratcode.parser.Token;
import be.jeffcheasey88.peeratcode.parser.TokenType;
import be.jeffcheasey88.peeratcode.parser.TokenValidator;
import be.jeffcheasey88.peeratcode.parser.Tokenizer;
import be.jeffcheasey88.peeratcode.parser.state.RedirectStateTree;
import be.jeffcheasey88.peeratcode.parser.state.StateTree;

public class JavaParser  extends Parser<JavaFile> {
	
	public static long time;
	
	public static void main(String[] args) throws Exception{
		File file = new File("C:\\Users\\jeffc\\eclipse-workspace\\peer-at-code-backend\\src\\be\\jeffcheasey88\\peeratcode\\routes\\Result.java");
		BufferedReader reader = new BufferedReader(new FileReader(file));
		
		time = System.currentTimeMillis();
		
		Parser<JavaFile> parser = new JavaParser();
		JavaFile jFile = new JavaFile();
		parser.parse(reader, jFile);

		System.out.println((System.currentTimeMillis()-time)+"ms");
	}
	
	private static final
		Function<TokenValidator, Boolean>
		LAMBDA_name = (v) -> v.validate((t) -> t.getType().equals(TokenType.NAME), (bag, token) -> {
			Token base = bag.get();
			if(base == null) bag.set(token);
			else bag.set(base.concat(token));
		}),
//		LAMBDA_name = (v) -> v.validate((t) -> {if(t.getType().equals(TokenType.NAME) && JavaParser.getModifier(t.getValue()) == 0) { System.out.println("valid type name --> "+t.getValue()); return true;}else return false;}),
		LAMBDA_mod = (v) -> v.validate((t) ->
			JavaParser.getModifier(t.getValue()) > 0,
				(bag, token) -> {
					Integer modifier = bag.get();
					if(modifier == null) modifier = 0;
					bag.set(modifier+JavaParser.getModifier(token.getValue()));
		});
		BiFunction<JavaFile, TokenValidator, JavaFile>
		END_PACKAGE = ((javafile, validator) -> javafile.setPackage(validator.getBag())),
		END_IMPORT = ((javafile, validator) -> javafile.addImport(validator.getBag()));
		BiFunction<JavaFile, TokenValidator, Class>
		END_CLASS = ((javafile, validator) -> javafile.setClass(new Class(validator.getBag())));
		BiFunction<JavaElement, TokenValidator, Token>
		END_TYPE = ((javafile, validator) -> {
			System.out.println("return token "+validator.getBag().get());
			return validator.getBag().get();
		});
		BiFunction<JavaElement, TokenValidator, Variable>
		END_VAR = ((javafile, validator) -> {
			System.out.println("new var");
			return new Variable();
		});
		BiFunction<JavaElement, TokenValidator, be.jeffcheasey88.peeratcode.parser.java.Function>
		END_FUNC = ((javafile, validator) -> {
			System.out.println("new function");
			return new be.jeffcheasey88.peeratcode.parser.java.Function(validator.getBag());
		});
		BiFunction<JavaElement, TokenValidator, Integer>
		END_MODIFIER = ((javafile, validator) ->{
			return validator.getBag().get();
		});

	public JavaParser(){
		Tokenizer tokenizer = new Tokenizer();

		Function<TokenValidator, Boolean> LAMBDA_0 = (v) -> v.validate((t) -> t.getValue().equals("package"));
		Function<TokenValidator, Boolean> LAMBDA_1 = (v) -> v.validate((t) -> t.getValue().equals(";"));
		Function<TokenValidator, Boolean> LAMBDA_3 = (v) -> v.validate((t) -> t.getValue().equals("import"));
		Function<TokenValidator, Boolean> LAMBDA_4 = (v) -> v.validate((t) -> t.getValue().equals("static"), (bag, token) -> bag.set("static", "bool"));
		Function<TokenValidator, Boolean> LAMBDA_5 = (v) -> v.validate((t) -> t.getValue().equals("<"));
		Function<TokenValidator, Boolean> LAMBDA_6 = (v) -> v.validate((t) -> t.getValue().equals(">"));
		Function<TokenValidator, Boolean> LAMBDA_7 = (v) -> v.validate((t) -> t.getValue().equals(","));
		Function<TokenValidator, Boolean> LAMBDA_8 = (v) -> v.validate((t) -> t.getValue().equals("{value}"));
		Function<TokenValidator, Boolean> LAMBDA_9 = (v) -> v.validate((t) -> t.getValue().equals("="));
		BiFunction<JavaElement, TokenValidator, JavaElement> END_10 = ((javaelement, validator) -> null);
		Function<TokenValidator, Boolean> LAMBDA_11 = (v) -> v.validate((t) -> t.getValue().equals("."));
		Function<TokenValidator, Boolean> LAMBDA_12 = (v) -> v.validate((t) -> t.getValue().equals("if"));
		Function<TokenValidator, Boolean> LAMBDA_13 = (v) -> v.validate((t) -> t.getValue().equals("("));
		Function<TokenValidator, Boolean> LAMBDA_14 = (v) -> v.validate((t) -> t.getValue().equals(")"));
		Function<TokenValidator, Boolean> LAMBDA_15 = (v) -> v.validate((t) -> t.getValue().equals("for"));
		Function<TokenValidator, Boolean> LAMBDA_16 = (v) -> v.validate((t) -> t.getValue().equals("while"));
		Function<TokenValidator, Boolean> LAMBDA_17 = (v) -> v.validate((t) -> t.getValue().equals("return"));
		Function<TokenValidator, Boolean> LAMBDA_18 = (v) -> v.validate((t) -> t.getValue().equals("synchronized"));
		Function<TokenValidator, Boolean> LAMBDA_19 = (v) -> v.validate((t) -> t.getValue().equals("continue"));
		Function<TokenValidator, Boolean> LAMBDA_20 = (v) -> v.validate((t) -> t.getValue().equals("throw"));
		Function<TokenValidator, Boolean> LAMBDA_21 = (v) -> v.validate((t) -> t.getValue().equals("break"));
		Function<TokenValidator, Boolean> LAMBDA_22 = (v) -> v.validate((t) -> t.getValue().equals("try"));
		Function<TokenValidator, Boolean> LAMBDA_23 = (v) -> v.validate((t) -> t.getValue().equals("do"));
		Function<TokenValidator, Boolean> LAMBDA_24 = (v) -> v.validate((t) -> t.getValue().equals("{"));
		Function<TokenValidator, Boolean> LAMBDA_25 = (v) -> v.validate((t) -> t.getValue().equals("}"));
		Function<TokenValidator, Boolean> LAMBDA_28 = (v) -> v.validate((t) -> t.getValue().equals("throws"));
		Function<TokenValidator, Boolean> LAMBDA_30 = (v) -> v.validate((t) -> t.getValue().equals("class"));
		Function<TokenValidator, Boolean> LAMBDA_31 = (v) -> v.validate((t) -> t.getValue().equals("implements"));
		Function<TokenValidator, Boolean> LAMBDA_32 = (v) -> v.validate((t) -> t.getValue().equals("extends"));

		StateTree<JavaFile> main = new StateTree<>();
		StateTree<JavaFile> package_q0 = new StateTree<>();
		package_q0.then(LAMBDA_0).then((v) -> v.validate((t) -> !t.getValue().equals(";"), groupToken("package"))).loop().then(LAMBDA_1).<JavaFile>end(END_PACKAGE);

		StateTree<JavaFile> import_q0 = new StateTree<>();
		StateTree<JavaFile> import_q1 = import_q0.then(LAMBDA_3);
		StateTree<JavaFile> import_q3 = import_q1.then((v) -> v.validate((t) -> !t.getValue().equals(";"), groupToken("import"))).loop();
		import_q3.then(LAMBDA_1).<JavaFile>end(END_IMPORT);
		StateTree<JavaFile> import_q2 = import_q1.then(LAMBDA_4);
		import_q2.then(import_q3);

		StateTree<JavaElement> modifier_q0 = new StateTree<>();
		modifier_q0.then(LAMBDA_mod).<Integer>end(END_MODIFIER);

		StateTree<JavaElement> type_q0 = new StateTree<>();
		StateTree<JavaElement> type_q1 = type_q0.then(LAMBDA_name);
		StateTree<JavaElement> type_q3 = type_q1.then(LAMBDA_5);
		StateTree<JavaElement> type_q4 = type_q3.then(LAMBDA_name);
		StateTree<JavaElement> type_q5 = type_q4.then(LAMBDA_6).loop();
		type_q5.<Token>end(END_TYPE);
		StateTree<JavaElement> type_q6 = type_q4.then(LAMBDA_7);
		type_q5.then(type_q6);
		type_q5.then(type_q4);
		type_q6.then(type_q4);
		type_q4.then(type_q3);
		type_q1.<Token>end(END_TYPE);

		StateTree<JavaElement> variable_q0 = new StateTree<>();
		StateTree<JavaElement> variable_q1 = variable_q0.then(new RedirectStateTree<>(type_q0,(bag) -> "type"));
		StateTree<JavaElement> variable_q2 = variable_q1.then(LAMBDA_name);
		StateTree<JavaElement> variable_q5 = variable_q2.then(LAMBDA_7);
		StateTree<JavaElement> variable_q7 = variable_q2.then(LAMBDA_9).then(LAMBDA_8);
		variable_q7.then(variable_q5);
		variable_q2.then(LAMBDA_1).<Variable>end(END_VAR);
		variable_q7.then(LAMBDA_1).<Variable>end(END_VAR);
		variable_q5.then(variable_q2);
		StateTree<JavaElement> variable_q10 = variable_q0.then(new RedirectStateTree<>(modifier_q0,(bag) -> "modifier"));
		variable_q10.then(variable_q1);

		StateTree<JavaElement> operation_q0 = new StateTree<>();
		StateTree<JavaElement> operation_q1 = operation_q0.then(LAMBDA_name);
		operation_q1.then(LAMBDA_9).then(LAMBDA_8).<JavaElement>end(END_10);
		StateTree<JavaElement> operation_q5 = operation_q1.then(LAMBDA_11);
		operation_q5.then(operation_q1);
		operation_q0.then(LAMBDA_12).then(LAMBDA_13).then(LAMBDA_8).then(LAMBDA_14).<JavaElement>end(END_10);
		operation_q0.then(LAMBDA_15).then(LAMBDA_13).then(LAMBDA_8).then(LAMBDA_14).<JavaElement>end(END_10);
		operation_q0.then(LAMBDA_16).then(LAMBDA_13).then(LAMBDA_8).then(LAMBDA_14).<JavaElement>end(END_10);
		operation_q0.then(LAMBDA_17).then(LAMBDA_8).then(LAMBDA_1).<JavaElement>end(END_10);
		operation_q0.then(LAMBDA_18).then(LAMBDA_13).then(LAMBDA_8).then(LAMBDA_14).<JavaElement>end(END_10);
		operation_q0.then(LAMBDA_19).then(LAMBDA_1).<JavaElement>end(END_10);
		operation_q0.then(LAMBDA_20).then(LAMBDA_8).then(LAMBDA_1).<JavaElement>end(END_10);
		operation_q0.then(LAMBDA_21).then(LAMBDA_1).<JavaElement>end(END_10);
		operation_q0.then(LAMBDA_22).<JavaElement>end(END_10);

		StateTree<be.jeffcheasey88.peeratcode.parser.java.Function> function_q17 = new StateTree<>();
		StateTree<be.jeffcheasey88.peeratcode.parser.java.Function> function_q18 = function_q17.then(LAMBDA_24);
		function_q18.then(LAMBDA_25).<Object>end((a,b) -> null);
		StateTree<be.jeffcheasey88.peeratcode.parser.java.Function> function_q22 = function_q18.then(new RedirectStateTree<>(operation_q0,(bag) -> "operation"));
		function_q22.then(function_q18);
		StateTree<be.jeffcheasey88.peeratcode.parser.java.Function> function_q21 = function_q18.then(new RedirectStateTree<>(variable_q0,(bag) -> "variable"));
		function_q21.then(function_q18);
		StateTree<JavaElement> function_q0 = new StateTree<>();
		StateTree<JavaElement> function_q1 = function_q0.then(new RedirectStateTree<>(type_q0,(bag) -> "return"));
		function_q1.then(LAMBDA_24).<be.jeffcheasey88.peeratcode.parser.java.Function>end(END_FUNC).then(function_q17);
		StateTree<JavaElement> function_q3 = function_q1.then((v) -> v.validate(
				(t) -> t.getType().equals(TokenType.NAME),
				(bag, token) -> bag.set("name", token))).then(LAMBDA_13);
		StateTree<JavaElement> function_q4 = function_q3.then(LAMBDA_14);
		StateTree<JavaElement> function_q10 = function_q4.then(LAMBDA_28).then(new RedirectStateTree<>(type_q0,(bag) -> "type"));
		function_q10.<be.jeffcheasey88.peeratcode.parser.java.Function>end(END_FUNC).then(function_q17);
		StateTree<JavaElement> function_q11 = function_q10.then(LAMBDA_7);
		function_q11.then(function_q10);
		function_q4.<be.jeffcheasey88.peeratcode.parser.java.Function>end(END_FUNC).then(function_q17);
		StateTree<JavaElement> function_q7 = function_q3.then(new RedirectStateTree<>(type_q0,(bag) -> "type")).then(LAMBDA_name);
		function_q7.then(function_q4);
		StateTree<JavaElement> function_q6 = function_q3.then(new RedirectStateTree<>(type_q0,(bag) -> "type"));
		StateTree<JavaElement> function_q8 = function_q7.then(LAMBDA_7);
		function_q8.then(function_q6);
		StateTree<JavaElement> function_q15 = function_q0.then(new RedirectStateTree<>(modifier_q0,(bag) -> "modifier"));
		function_q15.then(function_q1);

		StateTree<Class> class_q16 = new StateTree<>();
		StateTree<Class> class_q18 = class_q16.then(new RedirectStateTree<>(variable_q0,(bag) -> "?"));
		class_q18.then(class_q16);
		StateTree<Class> class_q19 = class_q16.then(new RedirectStateTree<>(function_q0,(bag) -> "?"));
		class_q19.then(class_q16);
		StateTree<JavaFile> class_q0 = new StateTree<>();
		StateTree<JavaFile> class_q1 = class_q0.then(LAMBDA_30);
		StateTree<JavaFile> class_q2 = class_q1.then(new RedirectStateTree<>(type_q0,(bag) -> "name"));
		StateTree<JavaFile> class_q9 = class_q2.then(LAMBDA_31);
		StateTree<JavaFile> class_q11 = class_q9.then(new RedirectStateTree<>(type_q0,(bag) -> "implement"));
		class_q11.then(LAMBDA_24).<Class>end(END_CLASS).then(class_q16);
		StateTree<JavaFile> class_q12 = class_q11.then(LAMBDA_7);
		class_q12.then(class_q11);
		StateTree<JavaFile> class_q7 = class_q2.then(LAMBDA_32).then(new RedirectStateTree<>(type_q0,(bag) -> "extend"));
		class_q7.then(class_q9);
		class_q7.then(LAMBDA_24).<Class>end(END_CLASS).then(class_q16);
		class_q2.then(LAMBDA_24).<Class>end(END_CLASS).then(class_q16);
		StateTree<JavaFile> class_q3 = class_q0.then(new RedirectStateTree<>(modifier_q0,(bag) -> "modifier"));
		class_q3.then(class_q1);

		main.then(new RedirectStateTree<>(package_q0, (b) -> null)).then(main);
		main.then(new RedirectStateTree<>(import_q0, (b) -> null)).then(main);
		main.then(new RedirectStateTree<>(class_q0, (b) -> null)).then(main);


		
		System.out.println((System.currentTimeMillis()-time)+"ms");
		
//		try {
//			DebugServer server = new DebugServer(main);
//		} catch (Exception e) {
//			e.printStackTrace();
//		}
		
		setTokenizer(tokenizer);
		setStateTree(main);
	}
	
	private BiConsumer<Bag, Token> groupToken(String group){
		return (bag, token) -> {
			Token base = bag.get(group);
			if(base == null) bag.set(group, token);
			else bag.set(group, base.concat(token));
		};
	}
	
	public static int getModifier(String modifier){
		switch(modifier){
			case "public": return Modifier.PUBLIC;
			case "private": return Modifier.PRIVATE;
			case "protected": return Modifier.PROTECTED;
			case "static": return Modifier.STATIC;
			case "final": return Modifier.FINAL;
			case "synchronized": return Modifier.SYNCHRONIZED;
			case "volatile": return Modifier.VOLATILE;
			case "transient": return Modifier.TRANSIENT;
			case "native": return Modifier.NATIVE;
			case "abstract": return Modifier.ABSTRACT;
			case "strictfp": return Modifier.STRICT;
			default: break;
		}
		return 0;
	}
	
}
